// Copyright (C) 2009 - Digiteo - Michael Baudin
// Copyright (C) INRIA - Farid BELAHCENE
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt

//
// number_gcd --
//   Returns the greatest common divisor of m and n
// Calling sequences
//   result = number_gcd ( m , n )
//   result = number_gcd ( m , n , method )
//
function result = number_gcd ( varargin )
  [lhs,rhs]=argn();
  if ( ( rhs <> 2 ) & ( rhs <> 3 ) ) then
    errmsg = msprintf(gettext("%s: Unexpected number of input arguments : %d provided while %d or %d are expected."), "number_gcd" , rhs , 2 , 3 );
    error(errmsg)
  end
  select rhs
  case 2 then
    m = varargin ( 1 )
    n = varargin ( 2 )
    method = "euclidian";
  case 3 then
    m = varargin ( 1 )
    n = varargin ( 2 )
    method = varargin ( 3 )
  end
  number_check ( m )
  number_check ( n )
  // gcd ( a , b ) = gcd ( -a , b ) = gcd ( a , -b)
  // therefore, take into account only for positive values.
  if ( m < 0 ) then
    m = -m
  end
  if ( n < 0 ) then
    n = -n
  end
  select method
  case "euclidian"
    result = gcd_euclidian ( m , n )
  case "brute"
    result = gcd_brute ( m , n )
  case "binary"
    result = gcd_binary ( m , n )
  case "euclidianiterative"
    result = gcd_euclidianiterative ( m , n )
  case "binaryiterative"
    result = gcd_binaryiterative ( m , n )
  else
    errmsg = sprintf ( gettext ( "%s: Unknown method %s" ) , "number_gcd" , method )
    error(errmsg)
  end
endfunction
//
// gcd_brute --
//   Returns the greatest common divisor of m and n
//   Use a brute-force method.
// Arguments:
//   m, n : the positive integer numbers
//
function result = gcd_brute ( m , n )
    if ( m > n ) then
      c = n
    else
      c = m
    end
    while ( %t )
      r1 = modulo ( m , c )
      r2 = modulo ( n , c )
      if ( ( r1==0 ) & ( r2==0 ) ) then
        break
      end
      c = c - 1
    end
    result = c
endfunction
//
// gcd_euclidian --
//   Returns the greatest common divisor of m and n
//   Use a naive recursive Euclidian method.
// Arguments:
//   m, n : the positive integer numbers
//
function result = gcd_euclidian ( m , n )
    if ( m==0 ) then
       result = n
       return
    end
    if ( n==0 ) then
       result = m
       return
    end
    if ( m>=n ) then
        r = modulo ( m , n )
        result = gcd_euclidian ( r , n )
    else
        r = modulo ( n , m )
        result = gcd_euclidian ( m , r )
    end
endfunction
//
// gcd_binary --
//   Returns the greatest common divisor of m and n
//   Use a naive binary method.
// Arguments:
//   m, n : the positive integer numbers
//
function result = gcd_binary ( m , n )
    if ( m==0 ) then
       result = n
       return
    end
    if ( n==0 ) then
       result = m
       return
    end
    meven = ( modulo ( m , 2 ) == 0 )
    neven = ( modulo ( n , 2 ) == 0 )
    if ( meven ) then
        if ( neven ) then
        // m even, n even
          m2 = m/2
          n2 = n/2
          d = gcd_binary ( m2 , n2 )
          result = 2 * d
          return
        else
        // m even, n odd
          m2 = m/2
          result = gcd_binary ( m2 , n )
          return
        end
     else 
        if ( neven ) then
        // m odd, n even
          n2 = n/2
          result = gcd_binary ( m , n2 )
          return
         else
        // m odd, n odd
          if ( m<=n ) then
            d = (n-m)/2
            result = gcd_binary ( m , d )
            return
          else
            d = (m-n)/2
            result = gcd_binary ( d , n )
            return
          end
        end
    end
endfunction
//
// gcd_euclidianiterative --
//   Returns the greatest common divisor of m and n
//   Use a naive, loop-based, Euclidian method.
// Arguments:
//   m, n : the positive integer numbers
//
  function result = gcd_euclidianiterative ( m , n )
    u = m
    v = n
    while ( u > 0 & v > 0 )
      if ( u >= v ) then
        u = modulo ( u , v )
      else
        v = modulo ( v , u )
      end
    end
    if ( u==0 ) then
      result = v
    elseif ( v==0 ) then
      result  = u
    else
      error(msprintf(gettext("%s: Unexpected case u = %d, v = %d"),"gcd_euclidianiterative" , u , v ));
    end
endfunction
//
// gcd_binaryiterative --
//   Returns the greatest common divisor of m and n
//   Use a naive, loop-based, binary method.
// Arguments:
//   m, n : the positive integer numbers
//
  function result = gcd_binaryiterative ( m , n )
    u = m
    v = n
  // functioness until u and v are odd,
  // or u or v is zero
    d = 1
    while ( u>0 & v>0 )
      ueven = ( modulo ( u , 2 ) ==0 )
      veven = ( modulo ( v , 2 ) ==0 )
      if ( ueven ) then 
        if ( veven ) then
        // u even, v even
          u = u/2 
          v = v/2 
          d = 2 * d
         else 
        // u even, v odd
          u = u/2
         end
      else
        if ( veven ) then
        // u odd, v even
          v = v/2
        else 
        // u odd, v odd
          if ( u<=v ) then
            v = (v-u)/2
          else 
            u = (u-v)/2
          end
        end
      end
    end
    if ( u==0 ) then
      result = d * v
    elseif ( v==0 ) then
      result = d * u
    else
      error(msprintf(gettext("%s: Unexpected case u = %d, v = %d"),"gcd_binaryiterative" , u , v ));
    end
endfunction

