// Copyright (C) 2009 - Digiteo - Michael Baudin
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt

//
// number_isprime --
//   Returns %t if the given number is prime.
//   If status is %t, then 
//   * isprime is true if n is prime.
//   * isprime is false if n is not prime.
//   If status is %f, then the primality test failed.
// Calling sequences
//   [ isprime , status ] = number_isprime ( n )
//   [ isprime , status ] = number_isprime ( n , method )
//
function varargout = number_isprime ( varargin )
  [lhs,rhs]=argn();
  if ( ( rhs <> 1 ) & ( rhs <> 2 ) ) then
    errmsg = msprintf(gettext("%s: Unexpected number of input arguments : %d provided while %d or %d are expected.\n"), "number_isprime" , rhs , 1 , 2 );
    error(errmsg)
  end
  if ( ( lhs <> 1 ) & ( lhs <> 2 ) ) then
    errmsg = msprintf(gettext("%s: Unexpected number of output arguments : %d provided while %d or %d are expected.\n"), "number_isprime" , lhs , 1 , 2 );
    error(errmsg)
  end
  select rhs
  case 1 then
    n = varargin ( 1 )
    method = "6k";
  case 2 then
    n = varargin ( 1 )
    method = varargin ( 2 )
  end
  number_check ( n )
  number_checkpositive ( "n" , n )
  imax = number_cget ( "-imax" )
  select method
  case "modulo"
    [ isprime , status ] = isprime_modulo ( n , imax )
  case "6k"
    [ isprime , status ] = isprime_6k ( n , imax )
  case "30k"
    [ isprime , status ] = isprime_30k ( n , imax )
  case "210k"
    [ isprime , status ] = isprime_210k ( n , imax )
  case "primes"
    [ isprime , status ] = isprime_primes ( n )
  case "massdivide"
    [ isprime , status ] = isprime_massdivide ( n )
  else
    errmsg = sprintf ( gettext ( "%s: Unknown method %s\n" ) , "number_isprime" , method )
    error(errmsg)
  end
  if ( ~status & ( lhs == 1 ) ) then
    mprintf(gettext("%s: Warning! The algorithm failed. Try to increase the value of the -imax option.\n"), "number_isprime" );
  end
  select lhs
  case 1 then
    varargout ( 1 ) = isprime
  case 2 then
    varargout ( 1 ) = isprime
    varargout ( 2 ) = status
  end
endfunction
//
// isprime_modulo --
//   Returns %t if the given number is prime.
//   Uses the modulo function.
//
function [ isprime , status ] = isprime_modulo ( n , imax )
    if ( n==1 ) then
      status = %t
      isprime = %f
      return
    end
    if ( n==2 ) then
      status = %t
      isprime = %t
      return
    end
    if ( modulo ( n , 2 ) ==0 ) then
      status = %t
      isprime = %f
      return
    end
    s = sqrt(n)
    d = 3
    verbose = number_cget ( "-verbose" )
    for i = 1 : imax
      if ( verbose > 0 ) then
        number_log ( sprintf ( "Loop #%d/%d\n" , i , imax ) )
        number_log ( sprintf ( "Testing divisor d=%s\n" , string(d) ) )
      end
      if ( d > s ) then
        number_log ( sprintf ( "All integers to test are greater than s=%s, n=%s is prime, done.\n", string(s) , string(n) ) )
        status = %t
        isprime = %t
        return
      end
      if ( modulo ( n , d ) ==0 ) then
        number_log ( sprintf ( "Number d=%s is a divisor of n=%s, n is not a prime, done.\n" , string(d) , string(n)) )
        status = %t
        isprime = %f
        return
      end
      d = d + 2
    end
    // We cannot say if the number is prime or not.
    status = %f
    isprime = %f
endfunction
//
// isprime_6k --
//   Returns %t if the given number is prime.
//   Based on modulo function and on the fact that prime 
//   numbers are of the form 6k + 1, 6k - 1
// Notes
//   The sequence of divisors to test is :
//   5 7
//   11 13
//   17 19
//   etc...
//
function [ isprime , status ] = isprime_6k ( n , imax )
    if ( n==1 ) then
      status = %t
      isprime = %f
      return
    end
    if ( or ( n == [ 2 3 ] ) ) then
      status = %t
      isprime = %t
      return
    end
    if ( or ( modulo ( n , [ 2 3 ] ) == 0 ) ) then
      status = %t
      isprime = %f
      return
    end
    s = sqrt(n)
    verbose = number_cget ( "-verbose" )
    number_log ( sprintf ( "Sqrt(n)=%s\n" , string(s) ) )
    for i = 1 : imax
      d = 6 * i + [-1 1]
      if ( verbose > 0 ) then
        number_log ( sprintf ( "Loop #%d/%d\n" , i , imax ) )
        number_log ( sprintf ( "Testing against divisors d=[%s]\n" , strcat ( string ( d ) , " " ) ) )
      end
      if ( and ( d > s ) ) then
        number_log ( sprintf ( "All integers to test are greater than s=%s, n=%s is a prime, done.\n", string(s) , string(n) ) )
        status = %t
        isprime = %t
        return
      end
      if ( or ( d > s ) ) then
        // Some integers are greater than s, but not all.
        // We must still check the integers which are lower or equal than s.
        number_log ( sprintf ( "Some integers to test are lower than s=%s, reducing the set of divisors.\n" , string(s) ) )
        indices = find ( d <= s )
        d = d ( indices )
        number_log ( sprintf ( "Reduced set is i=[%s]\n" , strcat ( string ( d ) , " " ) ) )
      end
      if ( or ( modulo ( n , d ) == 0 ) ) then
        number_log ( sprintf ( "Number n=%s is a multiple of some d=[%s], not prime, done.\n" , string(n) , strcat ( string ( d ) , " " ) ) )
        status = %t
        isprime = %f
        return
      end
   end
   // We cannot say if the number is prime or not.
   status = %f
   isprime = %f
endfunction
//
// isprime_primes --
//   Returns %t if the given number is prime.
//   Based on primes function
//
function [ isprime , status ] = isprime_primes ( n )
  pr = number_primes ( n )
  isprime = ( pr ( $ ) == n )
  status = %t
endfunction
//
// isprime_30k --
//   Returns %t if the given number is prime.
//   Based on modulo function and on the fact that prime 
//   numbers are of the form 30k + i, with i = 1, 7, 11, 13, 17, 19, 23, 29.
//   c#(5) = 2*3*5 = 30
//   The numbers 1,7,11,...,29 are the ones so that gcd(i,30)=1
//   We must first test for n = 2, 3 or 5.
//
function [ isprime , status ] = isprime_30k ( n , imax )
    if ( n==1 ) then
      status = %t
      isprime = %f
      return
    end
    if ( or ( n==[2 3 5] ) ) then
      status = %t
      isprime = %t
      return
    end
    if ( or ( modulo ( n , [ 2 3 5 ] ) ==0 ) ) then
      status = %t
      isprime = %f
      return
    end
    s = sqrt(n)
    verbose = number_cget ( "-verbose" )
    number_log ( sprintf ( "Sqrt(n)=%s\n" , string(s) ) )
    for i = 1 : imax
      d = 30 * i + [1 7 11 13 17 19 23 29]
      if ( verbose > 0 ) then
        number_log ( sprintf ( "Loop #%d/%d\n" , i , imax ) )
        number_log ( sprintf ( "Testing against divisors d=[%s]\n" , strcat ( string ( d ) , " " ) ) )
      end
      if ( and ( d > s ) ) then
        number_log ( sprintf ( "All integers to test are greater than s=%s, done.\n", string(s) ) )
        status = %t
        isprime = %t
        return
      end
      if ( or ( d > s ) ) then
        // Some integers are greater than s, but not all.
        // We must still check the integers which are lower or equal than s.
        number_log ( sprintf ( "Some integers to test are lower than s=%s, reducing the set of divisors.\n" , string(s) ) )
        indices = find ( d <= s )
        d = d ( indices )
        number_log ( sprintf ( "Reduced set is i=[%s]\n" , strcat ( string ( d ) , " " ) ) )
      end
      if ( or ( modulo ( n , d ) == 0 ) ) then
        number_log ( sprintf ( "Number n=%s is a multiple of some d=[%s], not prime, done.\n" , string(n) , strcat ( string ( d ) , " " ) ) )
        status = %t
        isprime = %f
        return
      end
    end
    // We cannot say if the number is prime or not.
    status = %f
    isprime = %f
endfunction
//
// isprime_massdivide --
//   Returns %t if the given number is prime.
//   Based on a massive division of n by all odd numbers up to sqrt(n).
//   This is fast, but requires all lot of memory if the number is large.
//
function [ isprime , status ] = isprime_massdivide ( n , imax )
    if ( n==1 ) then
      status = %t
      isprime = %f
      return
    end
    if ( or ( n==[2 3 5] ) ) then
      status = %t
      isprime = %t
      return
    end
    if ( or ( modulo ( n , 2 ) == 0 ) ) then
      status = %t
      isprime = %f
      return
    end
    s = sqrt(n)
    if ( s < 3 ) then
      status = %t
      isprime = %t
      return
    end
    q = n ./ ( 3 : 2 : s );
    isprime = and ( q <> int (q) );
    status = %t
endfunction

//
// isprime_210k --
//   Returns %t if the given number is prime.
//   Based on modulo function and on the fact that prime 
//   numbers are of the form 210k + i, with i = 1 11 13 17 ... 179 181 187 191 193 197 199 209 
//   We have c#(7) = 210.
//   The numbers 1,11,...,209  are the ones so that gcd(i,210)=1
//   We must first test for n = 2, 3 or 5 or 7 .
//
function [ isprime , status ] = isprime_210k ( n , imax )
    if ( n==1 ) then
      status = %t
      isprime = %f
      return
    end
    if ( or ( n==[2 3 5 7 ] ) ) then
      status = %t
      isprime = %t
      return
    end
    if ( or ( modulo ( n , [ 2 3 5 7 ] ) ==0 ) ) then
      status = %t
      isprime = %f
      return
    end
    s = sqrt(n)
    verbose = number_cget ( "-verbose" )
    number_log ( sprintf ( "Sqrt(n)=%s\n" , string(s) ) )
    table = [1 11 13 17 19 23 29 31 37 41 43 47 53 59 61 67 71 73 79 83 89 97 101 103 107 109 113 121 127 131 137 139 143 149 151 157 163 167 169 173 179 181 187 191 193 197 199 209 ]
    for i = 1 : imax
      d = 210 * i + table
      if ( verbose > 0 ) then
        number_log ( sprintf ( "Loop #%d/%d\n" , i , imax ) )
        number_log ( sprintf ( "Testing against divisors d=[%s]\n" , strcat ( string ( d ) , " " ) ) )
      end
      if ( and ( d > s ) ) then
        number_log ( sprintf ( "All integers to test are greater than s=%s, done.\n", string(s) ) )
        status = %t
        isprime = %t
        return
      end
      if ( or ( d > s ) ) then
        // Some integers are greater than s, but not all.
        // We must still check the integers which are lower or equal than s.
        number_log ( sprintf ( "Some integers to test are lower than s=%s, reducing the set of divisors.\n" , string(s) ) )
        indices = find ( d <= s )
        d = d ( indices )
        number_log ( sprintf ( "Reduced set is i=[%s]\n" , strcat ( string ( d ) , " " ) ) )
      end
      if ( or ( modulo ( n , d ) == 0 ) ) then
        number_log ( sprintf ( "Number n=%s is a multiple of some d=[%s], not prime, done.\n" , string(n) , strcat ( string ( d ) , " " ) ) )
        status = %t
        isprime = %f
        return
      end
    end
    // We cannot say if the number is prime or not.
    status = %f
    isprime = %f
endfunction

